;*****************************************************************************
; 
; Program	APTRECORDER.ASM
; Author	Wim de Vree, pe1grl@amsat.org
;	
; Revision history
;
;   2009-06-05	SPI memory support routines added using HW SSP
;		Added new fast FLASH calls, ISR for PWM audio playback/record
;   2009-06-04	Started work on the initial version
;   2009-07-30	"Simplified version", just record/playback a single message
;   2009-07-31	Implemented Sector Check/Erase + Button/Led
;   2009-08-01	Rewritten for 16 MHz Xtal (Will support 15625 Hz PWM with
;		10 bit PWM). Added audio-feedback on recording. Finished
;		Record/Playback routines
;   2009-08-02	Completed most of the stuff (async A/D, buffer flushing,
;		memory size test)
;   2009-08-03	Rewritten for 20Mhz Xtal (19531 Hz 10bit PWM)
;		Many 8 bit rates
;   2009-08-04	Functionality rewritten: Only record/playback while
;		the button is pressed. STOP button now is the ERASE
;		button, the other two should be switches.
;   2009-08-05	Multiple recordings can now be appended
;   2009-08-06	Adapted for Atmel 64Mbit chip
;   2009-08-08	Completed Atmel conversion, re-arranged I/O pins
;		Now uses Vref for a/d conversion
;   2009-08-09	ISR buffer enlarged from 64 to 80 bytes
;		VERSION 1.0 - 8 BIT SAMLING AT 9765 Hertz
;   2009-08-23	Modified for analog audio 2..4 Volt (2 Volt tt around 3 Volt)
;		Uses 9 bits of the A/D conversion
;		VERSION 1.1 - 8 BIT SAMLING AT 9765 Hertz
;=============================================================================

    	LIST P=PIC16F819, F=INHX8M
	__CONFIG _WDT_OFF & _HS_OSC & _LVP_OFF & _CCP1_RB3 & _MCLR_OFF & _PWRTE_ON

;#define	DIAGNOSTICS 	1	; Include diagnostic code (Serial I/O)
 
	    include <p16f819.inc>
	    include "../inc/macros2.inc"
	    include "../inc/m24bit.inc"

;*****************************************************************************
; Equates, I/O, vars
;*****************************************************************************
RESET_V	    	EQU	0x0000		        ; Address of RESET Vector
ISR_V		EQU	0x0004	        	; Address of Interrupt Vector
CLOCKFREQ	EQU	.200000			; 16 MHz

;----------------------------------------------------------------------
;           Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   InA In  Out Out In  X   X   X
;       B   Out In  Out Out In  In Out Out
;
; PortA pin0	IN	Audio input
;	pin1	IN	Not Used	
;	pin2	IN	Switch "RECORD"
;	pin3	IN	Vref for A/D (3 Volt)
;	pin4	IN  	Switch "PLAY"
;	pin5	IN  	Switch "ERASE"
;
; PortB pin0	OUT 	SPI Not Chip Select
;	pin1	IN  	SPI Data in (Q) 
;	pin2	OUT 	SPI Data out
;	pin3	OUT 	PWM Audio output
;	pin4	OUT 	SPI Clock
;	pin5	OUT	LED "RECORD"
;	pin6	OUT	LED "PLAY"
;	pin7	OUT	SERIAL OUTPUT (For diagnostic/debugging)
;----------------------------------------------------------------------

#define SW_RECORD       PORTA,2      	; Record audio
#define SW_PLAY		PORTA,4		; Playback audio
#define	SW_ERASE	PORTA,5		; Erase memory

#define	LED_REC		PORTB,5
#define	LED_PLAY	PORTB,6

;*****************************************************************************
; Variable definitions
;*****************************************************************************

	CBLOCK	0x20
	    SystemFlags				; Various status bits
  	    FM_Addr:0,FM_AddrLo,FM_AddrHi,FM_AddrBa ; Ext. 24bit memory address
	    FM_Data				; Data byte
	    FM_Temp				; Just a temp register
	    FM_Count				; Shift counter
	    WaveCnt:0,WaveCntLo,WaveCntMi,WaveCntHi; Counter used for wave playback
	    WaveSize:0,WaveSizeLo,WaveSizeMi,WaveSizeHi ; Length of wave to play
	    PWMValue:2				; Next value for PWM generator
            InsPtr,ExtPtr			; Ring buffer insert and extract pointer
	    FlashIdx				; Index in Flash Write Buffer
	    Sample
	    d1,d2,d3				; Used in Delay Routines
	    ADResult: 2				; Result of A/D conversion
	ENDC
#ifdef	DIAGNOSTICS
	CBLOCK					; Need some extra variables
	    B2A_RESULT: 3
	    SCRATCH: 8				; Working space for includes
	    Teller
	ENDC
#endif


	CBLOCK	0x7A			; Some allocations in "shared" memory
	    temp_S			; For use by Interrupt Service Routines
	    temp_W
	    temp_PCH
	    temp_FSR
	    GenCount			; Generic counter (always accesible)
	ENDC

	CBLOCK	0xA0			; Declare a 80 byte buffer in Bank1
	    Buffer:.80			; Used by ISR to store A/D samples
	ENDC

	CBLOCK	0x120			; Declare a 64 byte transfer buffer in Bank2
	    FlashBuf:.64		; used for Flash Write
	ENDC

#define	PlayBackMode		SystemFlags,1	; The PWM audio system is on
#define	RecordingMode		SystemFlags,2	; Recording is active

;*****************************************************************************
; Program start
;*****************************************************************************
	ORG	RESET_V			; RESET vector location
RESET		GOTO	START

;*****************************************************************************
; This is the Periperal Interrupt routine. Should NOT get here
;*****************************************************************************
	ORG	ISR_V			; Interrupt vector location
INTERRUPT	movwf   temp_W		; Save W and STATUS
		swapf   STATUS,W
	 	movwf   temp_S
		movfw	PCLATH
	 	movwf   temp_PCH
		clrf	PCLATH		; Select first 2K
	  	bcf     STATUS,RP0	; insure bank 0
		bcf	STATUS,IRP	; also for FSR

		; No need to check ISR source, we're only using TMR2
	   	bcf     PIR1,TMR2IF	; Clr Timer2 interrupt

i_CheckPlayBack	btfss	PlayBackMode	; PWM - Wave system enabled ?
		 goto	i_CheckRecord	; Nope

; PLAYBACK uses the hardware PWM module to output the data, the new sample
; is already available so write it to the PWM module and get the next value
; from the FLASH memory.

i_PlayBack				; Write Next PWM value to the hardware
		movfw	PWMValue	
		movwf	CCPR1L		; High 6 bits of PWM
		movfw	PWMValue+1
		movwf	CCP1CON		; Low  2 bits of PWM

		call	DecWaveCnt24	; Decrement wave counter
		skpnz			; End check for end of wave sample
		 goto	i_PlayBackDone

		incf	FM_AddrLo,F	; Increment address location
		skpnz
		 incf	FM_AddrHi,F
		skpnz
		 incf	FM_AddrBa,F

		call	FLASH_ReadNext
		call	SplitPWMValue
		goto	i_Done

i_PlayBackDone	bcf	T2CON, TMR2ON	; Stop TMR2
		bcf	PlayBackMode	; Stop Playback
		call	FLASH_DeSelect
		goto	i_Done

i_CheckRecord	btfss	RecordingMode	; Are we recording ?
		 goto	i_Done		; Nope

; When recording the ISR gets the result from the last A/D conversion
; an places it in a ring buffer. Then it starts a new A/D acquisition.
; The main code should take care of emptying the ringbuffer fast enough,
; the system halts if the buffer overflows

; Asynchronuous A/D conversion, we read the result from the last conversion
; and issue a new one. So we do not have to wait on de the conversion (20usec)

		btfsc	ADCON0,NOT_DONE	; Make sure conversion is complete
		goto	$-1

; Vref is 4 Volt, input is from 2 to 4 Volt, so the high bit is always set
; Test if the input voltage is below 2 Volt (MSB of ADRESH is 0), in that
; situation clip to zero to prevent garbage after the shift operation to
; get an 8 bit A/D value again (The actual conversion is 10 bit)

		btfss	ADRESH,7
		 goto	i_ADunderflow

		BANK01			; Bit 9&10 are in BANK1
		rlf	ADRESL,W	; 9'th bit to Carry
		BANK10
		rlf	ADRESH,W	; Now again an 8 bit result with
					; the MSB (always 1) removed
		iorlw	0		; RLF does not update Z
		goto	i_ADavailable

i_ADunderflow	andlw	0		; Use AND, not MOV. Need to set Z

i_ADavailable	bsf	ADCON0,GO	; Start a new conversion
		skpnz
		 movlw	1		; zero is not allowed for PWM
		movwf	ADResult	; Currently only 8 bit used

		movfw	ADResult	; Play Audio while recording
		call	SplitPWMValue
		movfw	PWMValue	
		movwf	CCPR1L		; High 6 bits of PWM
		movfw	PWMValue+1
		movwf	CCP1CON		; Low  2 bits of PWM

		call	DecWaveCnt24	; Decrement wave counter
		skpnz			; Check if memory is full
		 goto	i_RecordingDone

i_Add2Queue	movfw	FSR		; We need to use the FSR so save it
    		movwf	temp_FSR

		movfw	InsPtr      	; Get insert location
		addlw	Buffer		; Add start of ringbufer
	    	movwf	FSR		; FSR points to insert location (1st buffer)

		movfw	ADResult
	    	movwf	INDF		; store in ringbuffer1
	    	incf	InsPtr,F    	; Increment Extraction Pointer
		movlw	.80		; Wrap around after 80 bytes
		xorwf	InsPtr,W
		skpnz
		 clrf	InsPtr

   		movf	InsPtr,W	; check for buffer full
   		subwf	ExtPtr,W
	    	skpnz
	    	 goto	ShowError

		movfw	temp_FSR	; Restore FSR
    		movwf	FSR
		goto	i_Done

i_RecordingDone	bcf	T2CON, TMR2ON	; Stop TMR2
		bcf	RecordingMode	; Stop Recording

i_Done		movfw	temp_PCH	; Restore PCLATH, W and STATUS
		movwf	PCLATH
		swapf   temp_S,W
		movwf   STATUS
		swapf   temp_W,F
		swapf	temp_W,W
        	retfie
;----------------------------------------------------------------------------
; SplitPWMValue splits the 8bit value into a 6- and 2 bit part
; which are loaded into the PWM registers
;----------------------------------------------------------------------------
SplitPWMValue	movwf	PWMValue
		swapf	PWMValue,W
		andlw	b'00110000'
		iorlw	b'00001100'	; Set PWM mode, 1-channel
		movwf	PWMValue+1	; Low 2 bits of PWM
		rrf	PWMValue,F	; Get high 6 bits
		rrf	PWMValue,F
		movlw	0x3F
		andwf	PWMValue,F
		return

;----------------------------------------------------------------------------
; Decrement a 24bit wave counter and return Zero Status Yes/No
;----------------------------------------------------------------------------
DecWaveCnt24:	decfsz  WaveCntLo,F
        	 goto   _dec24nz

        	movfw   WaveCntMi
		iorwf	WaveCntHi,W	; Sets Zero status
		return

_dec24nz	clrz			; set _Z=0
		incfsz  WaveCntLo,W	; Rollover ?
		 return

		decfsz  WaveCntMi,F	; NOTE: Does not affect Zero status
		incfsz  WaveCntMi,W	; Rollover ?
		 return

		decfsz  WaveCntHi,F	; NOTE: Does not affect Zero status
		 return
		return

#ifdef	DIAGNOSTICS
	include "../inc/byt2asc.inc"
	include "../inc/byt2hex.inc"
#endif

; INCLUDE MEMORY SUPPORT ROUTINES, THIS IS FOR 4MBIT SPI FLASH
#define	USE_HARDWARE_SSP 1		; Use Hardware SPI

#define	SPI_NCS		PORTB, 0
#define	SPI_CLK		PORTB, 4
#define	SPI_D		PORTB, 2	; Data to SPI
#define	SPI_Q		PORTB, 1	; Data from SPI

;ST 512 4 Mbit chip
;	include	"M25PEXX.asm"		; Include support module for this chip

;Atmel AT45DB64D 64 Mbit chip with double buffer
	include	"AT45DB64D.asm"

#ifdef	DIAGNOSTICS
; INCLUDE SERIAL OUTPUT ROUTINE, THIS IS JUST FOR TESTING
#define	SERPORT		PORTB		; Port used for serial output
#define	TXBIT		7		; Bit used for serial output
#define	BDRATE		.9600		; requested Baud Rate
#define	MAXIM_USED	1		; MAX232 level converter used
	include <../inc/serout.inc>
#endif

;*****************************************************************************
; BEGIN OF MAIN CODE
;*****************************************************************************
START:					; POWER_ON Reset (Beginning of program)
		clrf	STATUS		; Do initialization, Select bank 0
		clrf	INTCON		; Clear int-flags, Disable interrupts
		clrf	PCLATH		; Keep in lower 2KByte

		clrf	PORTA		; ALL PORT output should output Low.
		clrf	PORTB

		clrf	ADCON0		; Disable comparators

		BANK1
		MOVLW	b'00011111'	; Configure I/O Ports see above
		MOVWF	TRISA
		movlw	b'00000010'
		movwf	TRISB
		BANK0

		movlw	b'10000001'	; Configure A/D converter
		movwf	ADCON0		; Fosc/32, channel 0, ON
		BANK1
		MOVLW	b'00001110'	; Left justified, 1 channel (AN0)
		movwf	ADCON1
		BANK0

		movlw   0x20		; Clear BANK0 memory
	        movwf   FSR
	        movlw	.80		; 80 bytes in BANK0 (excl. shared)
	        movwf	GenCount	; Clear them
_initMem	clrf    INDF
	        incf    FSR,F
	        decfsz	GenCount,f
	         goto	_initMem

#ifdef	DIAGNOSTICS
		call	SERINIT		; Init Serial Out interface
#endif
		call	FLASH_Init
		call	Delay200ms	; Small startup delay

		call	FLASH_Configure	; Needed for AT45DB642D one time init....
		skpc
		 goto	ShowError	; Power cycling is needed

#ifdef	DIAGNOSTICS
		call	Hello
		call	ShowMemoryID
		call	ShowStatus
		call	DumpMemory
		call	ShowSectors
;_tstloop	call	BlinkLeds
;		goto	_tstloop
#endif

		movlw	.5
		movwf	GenCount
StartBlink	call	BlinkLeds
		decfsz	GenCount,F
		 goto	StartBlink

MainLoop:	btfss	SW_RECORD
		 goto	StartRecording
		btfss	SW_PLAY
		 goto	StartPlayBack
		btfss	SW_ERASE
		 goto	StartEraseAll
		goto	MainLoop

StartRecording:	call	Delay20ms	; Debounce and removal of glitches
		btfsc	SW_RECORD
		 goto	MainLoop	; Just a glitch

		bsf	LED_REC

		call	GetAudioSize	; get Length of audio sample
		bnz	AppendRecording

NewRecording	clrf	WaveCntLo	; Complete FLASH is available
		clrf	WaveCntMi	; (AKA size of FLASH chip)
		movlw	((F_NRSECTORS*F_SECTORSIZE)/.65536)
		movwf	WaveCntHi
		call	DecWaveCnt24	; Decrement to allow "empty"

		CLR24	FM_Addr		; Reset Flash Address
		CLR24	WaveSize	; Wavefile length set to zero

		goto	_beginRecording

AppendRecording				; Appending to an existing sample
		call	AlignPage	; Always start at a new mem. page
		COPY24	WaveSize,FM_Addr

		clrf	WaveCntLo	; Set to size of flash
		clrf	WaveCntMi
		movlw	((F_NRSECTORS*F_SECTORSIZE)/.65536)
		movwf	WaveCntHi
					; Subtract already used part
		SUB24	WaveSize,WaveCnt
		call	DecWaveCnt24	; Decrement to allow "empty"

_beginRecording	clrf	InsPtr		; Empty Ringbuffer
		clrf	ExtPtr
		clrf	FlashIdx	; Empty Flash Buffer

		call	FLASH_StartSeqWrite ; Initiate writing at adres FM_Addr

#ifdef	DIAGNOSTICS
	        SNDCHAR 'R' 		; Recording start at 0x??????
        	SNDCHAR 'S' 
        	SNDCHAR '=' 
		movfw	FM_AddrBa
		call	SendHex
		movfw	FM_AddrHi
		call	SendHex
		movfw	FM_AddrLo
		call	SendHex
		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 
#endif
		bsf	RecordingMode
		bsf	ADCON0,GO	; Initiate first A/D conversion
		call	StartTMR2	; Here we go

; Recording principle: The ISR samples the data and places it in a
; ringbuffer. In the recording loop we extract the data from the
; ringbuffer and write it to FLASH if 64 bytes are sampled
RecordingLoop	btfsc	SW_RECORD	; Button still pressed ?
		 goto	StopRecording
		btfss	RecordingMode	; Still recording ?
		 goto	StopRecording	; maybe memory full
		

checkBuffer	movf	ExtPtr,W	; New (sampled) data available ?
    		subwf	InsPtr,W	; subtract extract-insert pointer
	    	skpnz
		 goto	RecordingLoop	; Buffer is empty, try again
		
		; Data available, retrieve it into the flash buffer

    		movf	ExtPtr,W	; Move Data from ring buffer to current value register
    		addlw	Buffer		; add start of Ringbuffer for recording
	    	movwf	FSR		; FSR points to extract location
	    	movf	INDF,W
	    	movwf	Sample		; Get Sample

	    	; Free the Ringbuffer entry, we have a copy in Sample
	    	; Note that the increment could be interrupted so always
	    	; keep the ptr valid !! (it is checked against in ISR)
		; This is why WREG is used in the increment
	
		incf	ExtPtr,W	; Increment Extraction Pointer
		xorlw	.80		; Wrap around after 80 bytes
		bz	_rl_wrap
_rl_nowrap	incf	ExtPtr,F	; No Wrap, increase Ptr
		goto	_rl_process
_rl_wrap	clrf	ExtPtr		; Wrap, reset Ptr
		
_rl_process	movfw	FlashIdx	; Place sample in FlashBuffer
		addlw	low FlashBuf	; And flush if needed
		movwf	FSR
	    	movfw	Sample
		bsf	STATUS,IRP	; Buffer is in Bank 2
	    	movwf	INDF		; Place in Flash Buffer
		bcf	STATUS,IRP	; Back to Bank 0/1

		incf	FlashIdx,F	; To next location
		btfss	FlashIdx,6	; Buffer holds 64 bytes ?
		 goto	RecordingLoop	; Nope

		; Buffer is full, write to FLASH memory
		call	FlushBuffer
		goto	RecordingLoop

StopRecording	bcf	T2CON, TMR2ON	; Stop Timer2
		bcf	RecordingMode

		movf	FlashIdx,F	; Data in Buffer ?
		skpz
		 call	FlushBuffer	; Yes, flush to EE
		call	FLASH_StopSeqWrite
		call	FMSize2EE	; Save for PlayBack/PowerDown

WaitRecSwitch	btfsc	SW_RECORD	; Button still pressed ?
		 goto	RecordingDone	; Wait for button to be released
		call	BlinkRecord
		goto	WaitRecSwitch

RecordingDone	call	Delay20ms
#ifdef	DIAGNOSTICS
	        SNDCHAR 'S' 
        	SNDCHAR 'Z' 
        	SNDCHAR '=' 
		movfw	FM_AddrBa
		call	SendHex
		movfw	FM_AddrHi
		call	SendHex
		movfw	FM_AddrLo
		call	SendHex
		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 

		call	DumpMemory	; Diagnostic
#endif

		bcf	LED_REC		; Led goes off
		goto	MainLoop

StartPlayBack:	call	Delay20ms	; Debounce and removal of glitches
		btfsc	SW_PLAY
		 goto	MainLoop	; Just a glitch

		bsf	LED_PLAY

#ifdef	DIAGNOSTICS
		call	GetAudioSize	; get Length of audio sample
		movwf	GenCount
	        SNDCHAR 'P' 
        	SNDCHAR 'B' 
        	SNDCHAR '=' 
		movfw	WaveSizeHi
		call	SendHex
		movfw	WaveSizeMi
		call	SendHex
		movfw	WaveSizeLo
		call	SendHex
		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 
		movwf	GenCount
#endif
		call	GetAudioSize	; get Length of audio sample
		bz	StopPlayBack	; Length == 0, nothing to play
					
		COPY24	WaveSize,WaveCnt; Set ISR counter for playback
		CLR24	FM_Addr		; Point to begin of FLASH memory

		call	FLASH_ReadSeq	; Read first sample
		call	SplitPWMValue	; Prepare CCP values
	
		bsf	PlayBackMode
		call	StartTMR2	; Here we go

PlayBackLoop				; Playback is completely ISR driven
					; So there is nothing in the loop
		btfsc	SW_PLAY		; Button still pressed ?
		 goto	StopPlayBack	; Nope, end playback
		btfss	PlayBackMode
		 goto	StopPlayBack	; End of sample reached
		goto	PlayBackLoop

StopPlayBack:	call	Delay20ms	; Debounce...
		bcf	T2CON, TMR2ON	; Stop Timer2
		bcf	PlayBackMode	; Stop Playback
		call	FLASH_DeSelect	; Make sure MEM is deselected

WaitPlaySwitch	btfsc	SW_PLAY		; Button still pressed ?
		 goto	PlaybackDone	; Wait for button to be released
		nop
		call	BlinkPlay
		goto	WaitPlaySwitch

PlaybackDone	call	Delay20ms
		bcf	LED_PLAY	; Clear Led
		goto	MainLoop


;----------------------------------------------------------------------------
; Align Page: An appended recording always starts at a new memory page
; so the size of the existing recording is rounded to a multiple of the
; memory page size
; WaveSize already contains the length of the existing recording(s)
AlignPage			
	movlw	low (F_PAGESIZE-1)	
	movwf	WaveCntLo		; Misuse this var.
	movlw	high (F_PAGESIZE-1)	
	movwf	WaveCntMi
	clrf	WaveCntHi
	ADD24	WaveCnt,WaveSize	; Adjust size

	movlw	low (F_PAGESIZE-1)	; Align to page size
	xorlw	0xFF
	andwf	WaveSizeLo,F
	movlw	high (F_PAGESIZE-1)
	xorlw	0xFF
	andwf	WaveSizeMi,F
	return

;----------------------------------------------------------------------------
; To Start erasing the button should be pressed down
; for at least .4 second (prevent accidental presses)
StartEraseAll:	call	Delay200ms
		call	Delay200ms
		btfsc	SW_ERASE	; Still pressed ?
		 goto	MainLoop	; No, do nothing

		bsf	LED_PLAY	; Show "Erase in Progress"
		bsf	LED_REC
#ifdef	DIAGNOSTICS
		call	ShowSectors
		bsf	LED_PLAY	; Might be turned of in diag code
		bsf	LED_REC
#endif
		call	FLASH_EraseAll
		CLR24	FM_Addr
		call	FMSize2EE
		bcf	LED_PLAY
		bcf	LED_REC

WaitEraseSwitch	btfsc	SW_ERASE
		 goto	MainLoop	; Switch is released
		call	BlinkLeds
		goto	WaitEraseSwitch

;----------------------------------------------------------------------------
; Write the uzed size of the FM memory to EE
FMSize2EE:	movlw	FM_Addr		; Write Length of audio to EE
		movwf	FSR
		movlw	.3		; 24 bits size (3bytes)
		movwf	GenCount
		movlw	low EE_AUDIOSIZE
		call	SaveEE
		return


;----------------------------------------------------------------------------
; Get the Length of the audiosample (fills in WaveSizeXX)
; Returns ZERO if no audio is stored (length == 0)

GetAudioSize	movlw	WaveSize	; Get Length of audio sample
		movwf	FSR		; FSR
		movlw	.3		; 24bits size
		movwf	GenCount
		movlw	low EE_AUDIOSIZE
		call	LoadEE

		movfw	WaveSizeLo	; Check for no sample stored
		iorwf	WaveSizeMi,W	; Sample Length == 0
		iorwf	WaveSizeHi,W
		return

;----------------------------------------------------------------------------
; FlushBuffer writes FlashIdx bytes to the SPI FLASH chip and adjusts
; the corresponding variables. We are using sequential Writing, so the
; Lowlevel code tracks the Flash Adress (FM_Addr)

FlushBuffer	movlw	low FlashBuf	; FSR -> Start of Buffer
		movwf	FSR		
		bsf	STATUS,IRP	; Buffer is in BANK2
		movfw	FlashIdx
		call	FLASH_SeqWrite64; Appends bytes
		bcf	STATUS,IRP	; Back to BANK0/1
		clrf	FlashIdx	; Buffer is empty again
		return

; **************************** PWM TIMER INITIALISE *************************
; ********************************* 8 BIT EXAMPLE ***************************
; PWM CARRIER is Xtal frequency / 256, the acquisition/replay rate is a
; factor N lower (N=1..15), XTAL = 20 MHz, PWM Carrier is 78125 Hz
; N=5: Recording/Playback at 15625   Hz
; N=6: Recording/Playback at 13020.8 Hz
; N=8: Recording/Playback at  9765.5 Hz
; N=9: Recording/Playback at  8680.6 Hz
; ******************************** 10 BIT EXAMPLE ***************************
; PWM CARRIER is Xtal frequency / 1024, the acquisition/replay rate is a
; factor N lower (N=1..15), XTAL = 20 MHz, PWM Carrier is 19531 Hz
; N=2: Recording/Playback at  9765.6 Hz

StartTMR2				; PWM system setup
;		movlw	b'00011000'	; Stopped, Pre 1:1, Post 1:4
;		movlw	b'00100000'	; Stopped, Pre 1:1, Post 1:5
;		movlw	b'00101000'	; Stopped, Pre 1:1, Post 1:6
;		movlw	b'00110000'	; Stopped, Pre 1:1, Post 1:7  11.161Hz
		movlw	b'00111000'	; Stopped, Pre 1:1, Post 1:8   9.765Hz
;		movlw	b'01000000'	; Stopped, Pre 1:1, Post 1:9   8.681Hz

		movwf	T2CON		
		clrf	TMR2		; Clear Timer2 register
		BANK1
		movlw	.63		; 0 -> 63
		movwf	PR2		; PWM carrier is 62.500 Hz
		BANK0
		movlw	.31		; 50 % duty cycle
		movwf	CCPR1L
		movlw	b'00111100'	; Enable PWM mode, lsb's are 1
		movwf	CCP1CON
	   	bcf     PIR1,TMR2IF	; Clr Timer2 interrupt
		BANK1
		bsf	PIE1,TMR2IE	; Timer 2 interrupt on 
		BANK0
		bsf	INTCON,PEIE	; Peripheral interrupts enabled
		bsf	INTCON,GIE	; Global interrupts enabled
		bsf	T2CON, TMR2ON	; Timer2 starts to increment
		return

;----------------------------------------------------------------------------
BlinkLeds:	
		bsf	LED_PLAY
		bsf	LED_REC
		call   	Delay200ms
		bcf	LED_PLAY
		bcf	LED_REC
		call   	Delay200ms
		return

BlinkPlay:	bsf	LED_PLAY
		call   	Delay200ms
		bcf	LED_PLAY
		call   	Delay200ms
		return

BlinkRecord:	bsf	LED_REC
		call   	Delay200ms
		bcf	LED_REC
		call   	Delay200ms
		return



;----------------------------------------------------------------------------
ShowError:	bcf	INTCON,GIE	; Hang system on error
		call	BlinkLeds	; with blinking leds
		goto	ShowError

#ifdef	DIAGNOSTICS
;----------------------------------------------------------------------------
SendHex		call	BYT2HEX
		movfw	SCRATCH+1
		movwf	B2A_RESULT
		movfw	SCRATCH+2
		movwf	B2A_RESULT+1
		movfw	B2A_RESULT
		call	SEROUT
		movfw	B2A_RESULT+1
		call	SEROUT
		movlw	' '
		call	SEROUT
		return

Hello:		SNDCHAR	0xFF
		SNDCHAR  'V'
		SNDCHAR  'o' 
        	SNDCHAR  'i' 
        	SNDCHAR  'c' 
        	SNDCHAR  'e' 
        	SNDCHAR  'R' 
       		SNDCHAR  'e' 
       		SNDCHAR  'c' 
		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 
        	return 

;----------------------------------------------------------------------------
ShowMemoryID:	call	FLASH_ReadID

	        SNDCHAR 'I' 
        	SNDCHAR 'D' 
        	SNDCHAR '=' 
		movfw	FM_AddrBa
		call	SendHex

	        SNDCHAR	'T' 
	        SNDCHAR	'P' 
	        SNDCHAR	'=' 
		movfw	FM_AddrHi
		call	SendHex

	        SNDCHAR	'S' 
	        SNDCHAR	'Z' 
	        SNDCHAR	'=' 
		movfw	FM_AddrLo
		call	SendHex

		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 
		return

ShowStatus:	SNDCHAR  'S' 
        	SNDCHAR  't' 
       		SNDCHAR  'a' 
       		SNDCHAR  't' 
       		SNDCHAR  'u' 
       		SNDCHAR  's' 
       		SNDCHAR  ' ' 
		call	FLASH_ReadStatus
		call	SendHex

		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 
		return

;----------------------------------------------------------------------------
ShowSectors:	CLR24	FM_Addr
		movlw	F_NRSECTORS
		movwf	GenCount
		clrf	Teller

_secLoop	SNDCHAR  'S' 
        	SNDCHAR  'e' 
       		SNDCHAR  'c' 
       		SNDCHAR  't' 
       		SNDCHAR  'o' 
       		SNDCHAR  'r' 
       		SNDCHAR  ' ' 
		movfw	Teller
		call	SendHex
       		SNDCHAR  ' ' 
;		bsf	LED_BUSY
		call	_CheckSector
;		bcf	LED_BUSY
		skpc
		 goto	ShowNotErased

ShowErased:    	SNDCHAR  'E' 
        	SNDCHAR  'r' 
       		SNDCHAR  'a' 
       		SNDCHAR  's' 
       		SNDCHAR  'e' 
       		SNDCHAR  'd' 
		goto	_secDone
     
ShowNotErased:	SNDCHAR  'I' 
        	SNDCHAR  'n' 
       		SNDCHAR  ' ' 
       		SNDCHAR  'U' 
       		SNDCHAR  's' 
       		SNDCHAR  'e' 

_secDone	SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 

		movlw	(F_SECTORSIZE>>.16) ; Should be multiple of 65K
		addwf	FM_AddrBa,F
		incf	Teller,F
		decfsz	GenCount,F
		 goto	_secLoop
		return

;----------------------------------------------------------------------------
; Dump the contents of the first 24 bytes of each 64K block
DumpMemory	CLR24	FM_Addr		
		movlw	((F_NRSECTORS*F_SECTORSIZE)/.65536)
		movlw	.8		; Only first 8 banks
		movwf	Teller	

_DumpLoop	movfw	FM_AddrBa
		call	SendHex
		SNDCHAR	'-'

		movlw	.24
		movwf	GenCount
		call	FLASH_ReadSeq

show24:		call	SendHex
		call	FLASH_ReadNext	; Single byte sequential
		decfsz	GenCount,F
		 goto	show24
		call	FLASH_DeSelect

		SNDCHAR  0x0D ; CR 
        	SNDCHAR  0x0A ; LF 

		incf	FM_AddrBa,F
		decfsz	Teller,F
		 goto	_DumpLoop
		return

;----------------------------------------------------------------------------
WriteTest:	movlw	Buffer
		movwf	FSR
		movlw	.32
		movwf	Teller
		movlw	0x20
		movwf	SCRATCH

_fbloop		movfw	SCRATCH
		movwf	INDF
		incf	FSR,F
		incf	SCRATCH,F
		decfsz	Teller,F
		 goto	_fbloop

		clrf	FM_AddrLo
		clrf	FM_AddrHi
		movlw	.5		
		movwf	FM_AddrBa	; Other part of memory for test

		movlw	Buffer
		movwf	FSR
;		bsf	LED_BUSY
		movlw	.32
		call	FLASH_ProgramMulti
;		bcf	LED_BUSY
		return

WriteSEQTest:	movlw	Buffer
		movwf	FSR
		movlw	.64
		movwf	Teller
		movlw	0x20
		movwf	SCRATCH

_fbloop2	movfw	SCRATCH
		movwf	INDF
		incf	FSR,F
		incf	SCRATCH,F
		decfsz	Teller,F
		 goto	_fbloop2

		clrf	FM_AddrLo
		clrf	FM_AddrHi
		movlw	0
		movwf	FM_AddrBa	; Other part of memory for test

		call	FLASH_StartSeqWrite

		movlw	.40
		movwf	Teller
_loop3		movlw	Buffer
		movwf	FSR
		call	FLASH_SeqWrite64
		decfsz	Teller,F
		 goto	_loop3

		return
#endif
;----------------------------------------------------------------------------

Delay20ms	movlw	0x1E		;99993 cycles
		movwf	d1
		movlw	0x4F
		movwf	d2
Delay20ms_0	decfsz	d1,F
		 goto	$+2
		decfsz	d2,F
		 goto	Delay20ms_0
		goto	$+1		;3 cycles
		nop
		return			;4 cycles (including call)

;----------------------------------------------------------------------------
; Delay = 0.2 seconds, Clock frequency = 20 MHz
; Actual delay = 0.2 seconds = 1.000.000 cycles

Delay200ms	movlw	0x08		;999997 cycles
		movwf	d1
		movlw	0x2F
		movwf	d2
		movlw	0x03
		movwf	d3
Delay200ms_0	decfsz	d1, f
		 goto	$+2
		 decfsz	d2, f
		 goto	$+2
		 decfsz	d3, f
		goto	Delay200ms_0
		goto	$+1	;3 cycles
		nop
		return			; 4 cycles (including call)

;----------------------------------------------------------------------------
; LoadEE loads COUNT nr of bytes from the EE.
; Upon entry:
;	GenCount	contains nr of bytes to load
;	W		contains starting adres in EE
;	FSR		points to destination location 
LoadEE		BANK02			; Bank 0->2 (EE registers are here)
		movwf	EEADR		; Load starting address
_readEE		BANK23
		bcf	EECON1,EEPGD	; Select DATA memory
		bsf	EECON1,RD	; Initiate EE read
		BANK32
		movfw	EEDATA		; Read byte
		movwf	INDF		; Store via FSR
		incf	FSR,f		; To next location in RAM
		incf	EEADR,f		; To next location in EE
		decfsz	GenCount,F
		 goto	_readEE
		BANK20
		return

;----------------------------------------------------------------------------
; SaveEE saves COUNT nr of bytes to the EE.
; Upon entry:
;	GenCount	contains nr of bytes to save
;	W		contains starting adres in EE
;	FSR		points to source location 
SaveEE		BANK02
		movwf	EEADR		; Set EE starting address
_writeEE	movfw	INDF		; Get byte to save
		movwf	EEDATA		; Write it
		BANK23
		bcf	EECON1,EEPGD	; Select DATA memory
		bsf	EECON1,WREN	; Enable EE Write
		bcf	INTCON,GIE	; Disable interrupts
		movlw	0x55		; Write required sequence
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf	EECON1,WR	; Initiate write cycle
		bsf	INTCON,GIE	; Enable interrupts again
		btfsc	EECON1,WR
		 goto	$-1		; Wait for write to complete (typ. 10msec)
		bcf	EECON1,WREN	; Disable EE Write
		BANK32
		incf	EEADR,f		; To next location in EE
		incf	FSR,f		; To next location in RAM
		decfsz	GenCount,F
		 goto	_writeEE
		BANK20
		return

		org 0x2100
EE_AUDIOSIZE
	de	0,0,0			; 24 bit size of stored audio 
	
	END
