;*****************************************************************************
;
; Include file for interfacing to a AT45DB65D Flash memory chip
; with an SPI interface. 
;
; Program	AT45DB65D.INC
; Author	Wim de Vree, pe1grl@amsat.org	
; Revision history
;	2009-08-05	Created 
;*****************************************************************************
;
; The following defines should be created in the main program and
; the datadirection setup properly
;#define	SPI_NCS		PORTA, 0	; NotChipSelect
;#define	SPI_CLK		PORTA, 1	; Clock		(SW)
;#define	SPI_D		PORTA, 3	; Data to SPI	(SW)
;#define	SPI_Q		PORTA, 4	; Data from SPI	(SW)
;
;The following variables should be declared:
;
;	FM_AddrBa		; Bank select (bits 23..16)
;	FM_AddrHi		; bits 15..8
;	FM_AddrLo		; bits  7..0
;	FM_Data			; Data byte
;	FM_Count		; Just a counter
;	FM_Temp			; Just a temp register
;
;USAGE:
;	FLASH_ErasePage		: Erases a 256 byte page
;	FLASH_Init		: Inits SPI datalines, no arguments
;	FLASH_ReadID		: Reads three ID bytes into FM_Addr
;	FLASH_Read		: reades a single byte, load FM_Addr with the
;				  starting address, returns the result in "W"
;

; LIST OF AVAILABLE COMMANDS TO THE CHIP
;#define	CMD_WREN	0x06		; Write Enable
;#define	CMD_WRDI	0x04		; Write Disable
#define	CMD_RDID	0x9F		; Read Identification
#define	CMD_RDSR	0xD7		; Read Status Register
#define	CMD_READ	0x03		; Read Data Bytes
#define	CMD_MM2B1	0x53		; Main Memory to Buffer 1
#define	CMD_WB1		0x84		; Write data to Buffer 1
#define	CMD_RB1		0xD1		; Read data from Buffer 1
#define	CMD_B12MM	0x88		; Program Buffer1 to Main Memory
#define	CMD_SE		0x7C		; Sector Erase

; STATUS FLAGS
#define	STS_BUSY	7		; Busy (Transfer/Program/Erase in progress)
#define	STS_PGSZ	0		; Device is in binary page mode

#define	F_NRSECTORS	.32		; Number of sectors in chip
#define	F_SECTORSIZE	.262144		; Size of each sector
#define	F_PAGESIZE	.1024		; Size of a single page

;----------------------------------------------------------------------------
FLASH_ReadID	bcf	SPI_NCS
		movlw	CMD_RDID
		call	_spiWriteByte
		call	_spiReadByte
		movwf	FM_AddrBa		; Manufacturer ID (0x20)
		call	_spiReadByte
		movwf	FM_AddrHi		; Memory type (0x80 for M25PE40)
		call	_spiReadByte
		movwf	FM_AddrLo		; Memory size (0x80 for M25PE40)
		bsf	SPI_NCS
		return

;----------------------------------------------------------------------------
FLASH_Read	bcf	SPI_NCS
		movlw	CMD_READ
		call	_spiWriteByte
		call	_spiWriteAddress
		call	_spiReadByte
		bsf	SPI_NCS
		return
	
;----------------------------------------------------------------------------
FLASH_ReadSeq	bcf	SPI_NCS
		movlw	CMD_READ
		call	_spiWriteByte
		call	_spiWriteAddress
		call	_spiReadByte
		return

;----------------------------------------------------------------------------
FLASH_ReadNext	goto	_spiReadByte

;----------------------------------------------------------------------------
FLASH_DeSelect	bsf	SPI_NCS
		return

#ifdef	USE_HARDWARE_SSP
;----------------------------------------------------------------------------
; Dedicated code to allow use the 1K buffer on this Atmel chip
; (feature only found on the double buffered variants)
; NOTE: we never use the second buffer
;----------------------------------------------------------------------------
; FLASH_StartSeqWrite: Load the buffer from the main memory
FLASH_StartSeqWrite
		bcf	SPI_NCS
		movlw	CMD_MM2B1	; Main memory --> Buffer 1 (1024bytes)
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady	; Wait until rdy (400 usec)
		return

;----------------------------------------------------------------------------
FLASH_SeqWrite64
		movlw	.64	
		movwf	FM_Data
		bcf	SPI_NCS
		movlw	CMD_WB1		; Data --> Buffer1
		call	_spiWriteByte
		call	_spiWriteAddress
_fswloop	movfw	INDF		; Get Data to write
		call	_spiWriteByte
		incf	FSR,F
		decfsz	FM_Data,F
		 goto	_fswloop
		bsf	SPI_NCS

		movlw	.64		; Adjust start adres
		addwf	FM_AddrLo,F
		skpz
		 return			; Not on multiple of 256
		movfw	FM_AddrHi
		andlw	0x03
		xorlw	0x03
		bz	_fsw_1K		; 1K Boundary

		incf	FM_AddrHi,F	; Not on a 1K boundary, just return
		skpnz
		 incf	FM_AddrBa,F
		return

_fsw_1K		bcf	SPI_NCS		; We have to copy the buffer to main mem.
		movlw	CMD_B12MM	; Buffer 1 --> Main Memory (1024 bytes)
		call	_spiWriteByte
		call 	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady	; Wait on completion

		incf	FM_AddrHi,F	; Now adjust the high part of the memory
		skpnz			; pointers
		 incf	FM_AddrBa,F

					; And get a new buffer for the current address
		bcf	SPI_NCS
		movlw	CMD_MM2B1	; Main memory --> Buffer 1 (1024bytes)
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady	; Wait until rdy (400 usec)

		return

;----------------------------------------------------------------------------
; If the FM_Addr pointer is not at a multiple of 1K then the buffer
; contains data which should be copied to the main memory
FLASH_StopSeqWrite
		movfw	FM_AddrLo
		skpz
		 goto	_SSW_1
		movfw	FM_AddrHi
		andlw	0xFC
		skpnz
		 return			; Nothing to do

_SSW_1		bcf	SPI_NCS
		movlw	CMD_B12MM	; Buffer 1 --> Main Memory
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady	; Wait until rdy (3 msec)
		return
#endif
	
;----------------------------------------------------------------------------
; ProgramMulti takes the same arguments as WriteMulti
; It can only write 1's to 0's zo the page should be erased before
; NOTE:	Do not wrap 256byte boundaries !!!
; NOTE: Very inefficient, is slow and stresses the chip.
; for each kByte 16 Write actions are used, it could also be
; done with a single Write !!!!

FLASH_ProgramMulti
		movwf	FM_Data
		bcf	SPI_NCS
		movlw	CMD_MM2B1	; Main memory --> Buffer 1 (1024bytes)
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady	; Wait until rdy (400 usec)

		bcf	SPI_NCS
		movlw	CMD_WB1		; Data --> Buffer1
		call	_spiWriteByte
		call	_spiWriteAddress
_fploop		movfw	INDF		; Get Data to write
		call	_spiWriteByte
		incf	FSR,F
		decfsz	FM_Data,F
		 goto	_fploop
		bsf	SPI_NCS

		bcf	SPI_NCS
		movlw	CMD_B12MM	; Buffer 1 --> Main Memory
		call	_spiWriteByte
		call 	_spiWriteAddress
		bsf	SPI_NCS
		call	_WaitReady
		return

;----------------------------------------------------------------------------
; EraseAll Erases the complete FLASH memory chip.
; First check, if a sector is already erased. Only issue an erase command
; if the sector contains values != 0xFF
; Also will set the audiosize to 0
;----------------------------------------------------------------------------
FLASH_EraseAll:	CLR24	FM_Addr
		movlw	F_NRSECTORS
		movwf	FM_Data

_EraseLoop	call	_CheckSector
		skpc			; If clear no need to Erase
		 call	FLASH_EraseSector
		
		movlw	(F_SECTORSIZE>>.16) ; Should be multiple of 65K
		addwf	FM_AddrBa,F
		decfsz	FM_Data,F
		 goto	_EraseLoop
		return

;----------------------------------------------------------------------------
; Check if a Sector is completely erased (containing 0xFF in every location)
; Should be called with the start adress of the sector in FM_Addr. As soom
; as we find a non-zero byte we can stop the scan. 
; Returns the status in the carry. C=1 if erased (all 0xFF), C=0 if in use
; Optimized with inline code to scan as fast as possible (Async. Reads)
; To get rid of 24bit math overhead we scan in 65K units
; DIRTY SPEEDUP FOR APTRECORDER: We just check the first 65K of each sector,
; if that is empty we assume the rest is also empty (The recorder writes
; sequential in the flash beginning at the start of memory)
;----------------------------------------------------------------------------
_CheckSector	clrf	FM_AddrLo	; Used in 65K counter
		clrf	FM_AddrHi
;		movlw	(F_SECTORSIZE>>.16)
		movlw	.1		; Just check first 65K
		movwf	FM_Temp		; No of 65K blocks in sector

		bcf	SPI_NCS
		movlw	CMD_READ
		call	_spiWriteByte
		call	_spiWriteAddress
		movlw	0xFF
		movwf	SSPBUF		; Start reading

_csloop		BANK01
		btfss	SSPSTAT,BF	; Wait for data to become available
		 goto	$-1
		BANK10
		movfw	SSPBUF		; Read received data
		xorlw	0xFF
		bnz	_SectorNotErased

		movlw	0xFF		; Initiate next byte
		movwf	SSPBUF		; Wether we need it or not

		movf   	FM_AddrLo,F	; 16bit counter using adres
  		skpnz			; If Low Byte == 0 Before 
   		 decf	FM_AddrHi,F	; Decrement High Byte
  		decf	FM_AddrLo,F	; Decrement Low Byte

		movfw	FM_AddrLo	; 65K Block done ?
		iorwf	FM_AddrHi,W
		bnz	_csloop

		decf	FM_Temp,F	; All Blocks done ?
		 bnz	_csloop

_SectorErased	BANK01			; There is an outstanding read
		btfss	SSPSTAT,BF	; Wait for completion
		 goto	$-1		
		BANK10
		movfw	SSPBUF		; Discard Data
		bsf	SPI_NCS		; Deselect and return the proper
		clrf	FM_AddrLo	; Reset again
		clrf	FM_AddrHi
		setc			; status
		return

_SectorNotErased			; No Outstanding reads, just deselect
		bsf	SPI_NCS		; the chip and return the proper status
		clrf	FM_AddrLo	; Reset again
		clrf	FM_AddrHi
		clrc
		return

;----------------------------------------------------------------------------
; Erases the sector using a given start address
FLASH_EraseSector
		bcf	SPI_NCS
		movlw	CMD_SE		; Erase Sector
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS		; deselect initiates the erase
		call	_WaitReady

		movfw	FM_AddrLo	; Check to see if it was sector0
		iorwf	FM_AddrHi,W	; which is special: split in an 0a
		iorwf	FM_AddrBa,W	; and a 0b part
		skpz	
		 return			; Sector 1..31

		movlw	0x20		; Sector 0b start at 8192
		movwf	FM_AddrHi
		bcf	SPI_NCS		; 2 Erases: Block 0a and Block 0b
		movlw	CMD_SE		; Erase Sector 0b
		call	_spiWriteByte
		call	_spiWriteAddress
		bsf	SPI_NCS		; deselect initiates the erase
		call	_WaitReady
		return

;----------------------------------------------------------------------------
FLASH_ReadStatus
		bcf	SPI_NCS		; enable chip
		movlw	CMD_RDSR
		call	_spiWriteByte	; Write Command
		call	_spiReadByte	; Get status
		bsf	SPI_NCS		; disable chip
		return	

;----------------------------------------------------------------------------
; FLASH_Configure: The Atmel AT64DB642D comes with a page size of 1056 bytes
;		   It should be reprogrammed to use a 1024 byte buffer, this
;		   is a one-time only option and should be executed once
;		   Returns with carry clear if a powerdown cycle is required
FLASH_Configure	call	FLASH_ReadStatus
		setc
		btfsc	FM_Temp,STS_PGSZ; In Binary page mode ?
		 return			; Yep, return

		bcf	SPI_NCS		; No, send sequence to configure
		movlw	0x3D		; for binary page size
		call	_spiWriteByte
		movlw	0x2A
		call	_spiWriteByte
		movlw	0x80
		call	_spiWriteByte
		movlw	0xA6
		call	_spiWriteByte
		call	_WaitReady
		clrc			; Power cycling required !!
		return
		
;----------------------------------------------------------------------------
; Wait for the result of an Erase, Write or Program Cycle
; by continuously polling the Status Register WIP bit
_WaitReady	bcf	SPI_NCS		; enable chip
		movlw	CMD_RDSR
		call	_spiWriteByte	; Write Command
_wwcloop	call	_spiReadByte	; Get status
		btfss	FM_Temp,STS_BUSY
		 goto	_wwcloop
		bsf	SPI_NCS		; disable chip
		return		

	include "F_SPI_IO.ASM"
