;*****************************************************************************
;
; Include file containing low level routines for SPI Flash memories.
; Can be configured for the hardware SSP or for bitbanged I/O.
;
; Program	FM_SPI_IO.INC
; Author	Wim de Vree, pe1grl@amsat.org	
; Revision history
;	2009-08-05	Seperated chip independend stuff to this file
;*****************************************************************************
;
; The following defines should be created in the main program and
; the datadirection setup properly
;#define	SPI_NCS		PORTA, 0	; NotChipSelect
;#define	SPI_CLK		PORTA, 1	; Clock		(SW)
;#define	SPI_D		PORTA, 3	; Data to SPI	(SW)
;#define	SPI_Q		PORTA, 4	; Data from SPI	(SW)
;
;The following variables should be declared:
;
;	FM_AddrBa		; Bank select (bits 23..16)
;	FM_AddrHi		; bits 15..8
;	FM_AddrLo		; bits  7..0
;	FM_Data			; Data byte
;	FM_Count		; Just a counter
;	FM_Temp			; Just a temp register
;
; Transfer Rates for a 20 MHz PIC (software version)
; FLASH_Read:		 87.4 usec		(11 kB/sec)
; FLASH_ReadMulti:	72.2 + n*16.8 usec	(n=32 52 kB/sec, n=64 56 kB/sec)
;
; Reading 64bytes with FLASH_ReadMulti
;	SOFTWARE LOOP:		2860 usec (8MHz)
;	SOFTWARE INLINE		2000 usec (8MHz)
;	HARDWARE SSP		 871 usec (8MHz)
;

;----------------------------------------------------------------------------
FLASH_Init	bsf	SPI_NCS
#ifdef	USE_HARDWARE_SSP
		BANK1
		movlw	b'11000000'	
		movwf	SSPSTAT
		BANK0
		movlw	b'00100000'	; SPI MASTER, Fosc/4
;		movlw	b'00100001'	; SPI MASTER, Fosc/16
;		movlw	b'00100010'	; SPI MASTER, Fosc/64
		movwf	SSPCON
#else
		bcf	SPI_CLK
		bcf	SPI_D
#endif
		return

#ifdef	USE_HARDWARE_SSP
;----------------------------------------------------------------------------
_spiWriteByte	movwf	SSPBUF
		BANK01
		btfss	SSPSTAT,BF	; Wait for completion
		 goto	$-1
		BANK10
		movfw	SSPBUF		; Dummy read
		return

;----------------------------------------------------------------------------
_spiReadByte	movlw	0xFF
		movwf	SSPBUF		; Dummy Write
		BANK01
		btfss	SSPSTAT,BF	; Wait for completion
		 goto	$-1
		BANK10
		movfw	SSPBUF		; Read received data
		movwf	FM_Temp
		return
#else
#ifdef	SPI_INLINE
;----------------------------------------------------------------------------
SPIPUTBIT	MACRO
		rlf	FM_Temp,F
		skpc
		 bcf	SPI_D		; Output "0"
		skpnc
		 bsf	SPI_D		; Output "1'
		bsf	SPI_CLK		; Clock up
		bcf	SPI_CLK		; Clock down
		ENDM

;----------------------------------------------------------------------------
SPIGETBIT	MACRO
		clrc			; Is Data IN high or low ?
		btfsc	SPI_Q		; set result in carry
		 setc
		rlf	FM_Temp,F	; Shift bit in TEMP
		bsf	SPI_CLK
		bcf	SPI_CLK
		ENDM

;----------------------------------------------------------------------------
_spiWriteByte	movwf	FM_Temp
		SPIPUTBIT		; Bit 7
		SPIPUTBIT		; Bit 6
		SPIPUTBIT		; Bit 5
		SPIPUTBIT		; Bit 4
		SPIPUTBIT		; Bit 3
		SPIPUTBIT		; Bit 2
		SPIPUTBIT		; Bit 1
		SPIPUTBIT		; Bit 0
		return

;----------------------------------------------------------------------------
_spiReadByte	clrf	FM_Temp
		SPIGETBIT		; Bit 7
		SPIGETBIT		; Bit 6
		SPIGETBIT		; Bit 5
		SPIGETBIT		; Bit 4
		SPIGETBIT		; Bit 3
		SPIGETBIT		; Bit 2
		SPIGETBIT		; Bit 1
		SPIGETBIT		; Bit 0
		movfw	FM_Temp		; Also place result in W
		return
#else
;----------------------------------------------------------------------------
_spiWriteByte	movwf	FM_Temp
		movlw	.8
		movwf	FM_Count
_spiWloop	rlf	FM_Temp,F
		skpc
		 bcf	SPI_D		; Output "0"
		skpnc
		 bsf	SPI_D		; Output "1'
		bsf	SPI_CLK		; Clock up
		bcf	SPI_CLK		; Clock down
		decfsz	FM_Count,F
		 goto	_spiWloop
		return

;----------------------------------------------------------------------------
_spiReadByte	clrf	FM_Temp
		movlw	.8
		movwf	FM_Count
_spiRloop	clrc			; Is Data IN high or low ?
		btfsc	SPI_Q		; set result in carry
		 setc
		rlf	FM_Temp,F
		bsf	SPI_CLK
		bcf	SPI_CLK
		decfsz	FM_Count,F
		 goto	_spiRloop
		movfw	FM_Temp		; Also place result in W
		return
#endif
#endif
;----------------------------------------------------------------------------
_spiWriteAddress			; Write 3 bytes
		movfw	FM_AddrBa
		call	_spiWriteByte
		movfw	FM_AddrHi
		call	_spiWriteByte
		movfw	FM_AddrLo
		call	_spiWriteByte
		return

;----------------------------------------------------------------------------
; Write an 8 bit command, no addres, no data
; Call with Command opcode in WREG
_spiWriteCmd	bcf	SPI_NCS		; enable chip
		call	_spiWriteByte	; Write Command
		bsf	SPI_NCS		; disable chip
		return
